/* beginnings of a PDF parser in hammer
 * pesco 2019
 */

#include <string.h>	/* strncmp(), memset() */
#include <stdlib.h>	/* exit() */

#include <hammer/hammer.h>
#include <hammer/glue.h>

/* convenience macros */
#define SEQ(...)	h_sequence(__VA_ARGS__, NULL)
#define CHX(...)	h_choice(__VA_ARGS__, NULL)
#define OPT(X)		h_ignore(h_optional(X))
#define REP(P,N)	h_repeat_n(P, N)
#define IGN(P)		h_ignore(P)
#define LIT(S)		h_literal(S)
#define IN(STR)		h_in((const uint8_t *)(STR), sizeof(STR) - 1)
#define NOT_IN(STR)	h_not_in((const uint8_t *)(STR), sizeof(STR) - 1)

#ifdef LOG
#define VIOL(P,VIOL)	 h_action(h_sequence(P, h_tell(), NULL), act_viol, VIOL)
#else
#define VIOL(P,VIOL)	 P
#endif



/*
 * some helpers
 */

HParser *p_fail;
HParser *p_epsilon;
HParser *p_return_0;
HParser *p_return_1;

uint8_t strictness = 0;

/* a combinator to parse a given character but return a different value */

HParsedToken *
act_return_uint(const HParseResult *p, void *u)
{
	return H_MAKE_UINT((uint64_t)(uintptr_t)u);
}

HParser *
p_mapch(uint8_t c, uint8_t v)
{
	return h_action(h_ch(c), act_return_uint, (void *)(uintptr_t)v);
}

/* a parser that just returns a given token */

HParsedToken *
act_return(const HParseResult *p, void *u)
{
	return u;
}

HParser *
p_return__m(HAllocator *mm__, const HParsedToken *tok)
{
	return h_action__m(mm__, p_epsilon, act_return, (void *)tok);
}

/* a parser that just returns a given uint */
HParser *
p_return_uint__m(HAllocator *mm__, uint64_t x)
{
	return h_action__m(mm__, p_epsilon, act_return_uint, (void *)x);
}

/* a helper to compare an HBytes to a string */
bool
bytes_eq(HBytes b, const char *s)
{
	return strncmp(s, (const char *)(b.token), b.len) == 0 && b.len == strlen(s);
}

bool
validate_eq_uint(HParseResult *p, void *u)
{
	const HParsedToken *v = p->ast;
	return (v != NULL && v->token_type == TT_UINT &&
	    v->uint == (uint64_t)(uintptr_t)u);
}

/*
 * auxiliary global data structure needed by the parser
 */
struct Env {
	const char *infile;
	const uint8_t *input;
	size_t sz;

	const HParsedToken **xrefs;	/* all xref sections of the file */
	size_t nxrefs;
};


/*
 * custom token types
 */
HTokenType TT_XREntry, TT_Ref, TT_Dict, TT_HParseResult;

typedef struct {
	enum {XR_FREE, XR_INUSE, XR_OBJSTM} type;
	union {
		struct { size_t next, ngen; } f;	/* free */
		struct { size_t offs, gen; } n;		/* inuse */
		struct { size_t stm, idx; } o;		/* objstm */
	};
	const HParsedToken *obj;
} XREntry;

typedef struct { size_t nr, gen; } Ref;

typedef HCountedArray Dict;

/* look up a value in a dictionary */
const HParsedToken *
dictentry(const Dict *dict, const char *key)
{
	HParsedToken *ent;
	HBytes k;
	size_t len;

	if (dict == NULL)
		return NULL;

	len = strlen(key);
	for (size_t i = 0; i < dict->used; i++) {
		ent = dict->elements[i];
		k = H_INDEX_BYTES(ent, 0);

		if (k.len == len && bytes_eq(k, key))
			return H_INDEX_TOKEN(ent, 1);
	}

	return NULL;
}

void
pp_xrentry(FILE *stream, const HParsedToken *tok, int indent, int delta)
{
	XREntry *xr = H_CAST(XREntry, tok);

	switch (xr->type) {
	case XR_FREE:
		fprintf(stream, "{ \"type\":\"free\", \"next\":%zu, "
		    "\"ngen\":%zu }", xr->f.next, xr->f.ngen);
		break;
	case XR_INUSE:
		fprintf(stream, "{ \"type\":\"inuse\", \"offs\":%zu, "
		    "\"gen\":%zu }", xr->n.offs, xr->n.gen);
		break;
	case XR_OBJSTM:
		fprintf(stream, "{ \"type\":\"objstm\", \"stm\":%zu, "
		    "\"idx\":%zu }", xr->o.stm, xr->o.idx);
		break;
	default:
		assert(!"reached");
	}
}

void
pp_ref(FILE *stream, const HParsedToken *tok, int indent, int delta)
{
	Ref *r = H_CAST(Ref, tok);

	fprintf(stream, "[%zu,%zu]", r->nr, r->gen);
}

void
pp_parseresult(FILE *stream, const HParsedToken *tok, int indent, int delta)
{
	HParseResult *res = H_CAST(HParseResult, tok);

	h_pprint(stream, res->ast, indent, delta);
}

void
pp_dict(FILE *stream, const HParsedToken *tok, int indent, int delta)
{
	const HCountedArray *k_v;

	if (tok->seq->used == 0) {
		fprintf(stream, "{ }");
		return;
	}

	fprintf(stream, "{%*s", delta - 1, "");
	for (size_t i = 0; i < tok->seq->used; i++) {
		k_v = tok->seq->elements[i]->seq;
		if (i > 0)
			fprintf(stream, "\n%*s,%*s", indent, "", delta - 1, "");
		h_pprint(stream, k_v->elements[0], indent + delta, delta);
		fprintf(stream, ":");
		h_pprint(stream, k_v->elements[1], indent + delta, delta);
	}

	if (tok->seq->used > 2)
		fprintf(stream, "\n%*s}", indent, "");
	else
		fprintf(stream, " }");
}

/*
 * semantic actions
 */

HParsedToken *
act_digit(const HParseResult *p, void *u)
{
	return H_MAKE_UINT(H_CAST_UINT(p->ast) - '0');
}
#define act_pdigit act_digit
#define act_odigit act_digit

HParsedToken *
act_hlower(const HParseResult *p, void *u)
{
	return H_MAKE_UINT(H_CAST_UINT(p->ast) - 'a');
}

HParsedToken *
act_hupper(const HParseResult *p, void *u)
{
	return H_MAKE_UINT(H_CAST_UINT(p->ast) - 'A');
}

HParsedToken *
act_nat(const HParseResult *p, void *u)
{
	uint64_t x = 0;
	HCountedArray *seq = H_CAST_SEQ(p->ast);

	// XXX check for overflow
	for (size_t i = 0; i < seq->used; i++)
	    x = x*10 + (H_CAST_UINT(seq->elements[i]));

	return H_MAKE_UINT(x);
}
#define act_xroff act_nat
#define act_xrgen act_nat

HParsedToken *
act_viol(const HParseResult *p, void *viol)
{
	uint severity;
	uint offset;
	viol = (uint8_t *) viol;
	H_RULE(viol_preamble, SEQ(h_many(NOT_IN("=")), LIT("=")));
	H_RULE(nat, h_action(h_many1(h_action(h_ch_range('0', '9'), act_digit, NULL)),
			     act_nat, NULL));
	H_RULE(p_violsev, SEQ(IGN(viol_preamble), nat));
	severity = h_parse(p_violsev, viol, strlen((char *)viol))->ast->seq->elements[0]->uint;
	offset = p->ast->seq->elements[1]->uint / 8;
	fprintf(stderr, "VIOLATION[%d]@%d (0x%x): %s\n", severity, offset, offset, (char *) viol);
	if (strictness && severity > strictness) {
	    exit(1);
	}
	/* Just return the parse AST, drop the h_tell */
	return (HParsedToken *) p->ast->seq->elements[0];
}

bool
validate_pnat(HParseResult *p, void *u)
{
	return H_CAST_UINT(p->ast) > 0;
}

HParsedToken *
act_sign(const HParseResult *p, void *u)
{
	return H_MAKE_SINT(p->ast == NULL ? 1 : -1);
}

/* p = (digits, period, digits) */
HParsedToken *
act_realnn(const HParseResult *p, void *u)
{
	double x = 0;
	double f = 0;
	HCountedArray *whole = H_FIELD_SEQ(0);
	HCountedArray *fract = H_FIELD_SEQ(2);

	// XXX check for overflow
	for (size_t i = 0; i < whole->used; i++)
		x = x*10 + H_CAST_UINT(whole->elements[i]);
	for (size_t i = 0; i < fract->used; i++)
		f = (f + H_CAST_UINT(fract->elements[fract->used - 1 - i])) / 10;

	return H_MAKE_DOUBLE(x + f);
}

bool
validate_intnn(HParseResult *p, void *u)
{
	return H_CAST_UINT(p->ast) <= INT64_MAX;
}

/* p = numbnn | (sign, numbnn) */
HParsedToken *
act_numb(const HParseResult *p, void *u)
{
	const HParsedToken *x = p->ast;
	int64_t sgn = 1;

	if (x->token_type == TT_SEQUENCE) {
		sgn = H_FIELD_SINT(0);
		x = H_FIELD_TOKEN(1);
	}

	assert(sgn == 1 || sgn == -1);
	switch (x->token_type) {
	case TT_UINT:
		return H_MAKE_SINT(sgn * x->uint);
	case TT_DOUBLE:
		return H_MAKE_DOUBLE(sgn * x->dbl);
	default:
		assert(!"reached");
	}
}

HParsedToken *
act_ref(const HParseResult *p, void *u)
{
	Ref *r = H_ALLOC(Ref);

	r->nr = H_FIELD_UINT(0, 0);
	r->gen = H_FIELD_UINT(0, 1);

	return H_MAKE(Ref, r);
}

HParsedToken *
act_token(const HParseResult *p, void *u)
{
	HCountedArray *seq = H_CAST_SEQ(p->ast);
	uint8_t *bytes;

	bytes = h_arena_malloc(p->arena, seq->used);
	for (size_t i = 0; i < seq->used; i++)
		bytes[i] = H_CAST_UINT(seq->elements[i]);

	return H_MAKE_BYTES(bytes, seq->used);
}
#define act_nstr act_token

HParsedToken *
act_nesc(const HParseResult *p, void *u)
{
	return H_MAKE_UINT(H_FIELD_UINT(1)*16 + H_FIELD_UINT(2));
}

#define act_schars h_act_flatten
#define act_string act_token

HParsedToken *
act_octal(const HParseResult *p, void *u)
{
	uint64_t x = 0;
	HCountedArray *seq = H_CAST_SEQ(p->ast);

	for (size_t i = 0; i < seq->used; i++)
		x = x*8 + H_CAST_UINT(seq->elements[i]);

	return H_MAKE_UINT(x);
}

HParsedToken *
act_xrent(const HParseResult *p, void *u)
{
	XREntry *xr = H_ALLOC(XREntry);
	char c = H_FIELD_UINT(2);

	xr->obj = NULL;
	switch (c) {
	case 'f':
		xr->type = XR_FREE;
		xr->f.next = H_FIELD_UINT(0);
		xr->f.ngen = H_FIELD_UINT(1);
		break;
	case 'n':
		xr->type = XR_INUSE;
		xr->n.offs = H_FIELD_UINT(0);
		xr->n.gen = H_FIELD_UINT(1);
		break;
	default:
		assert(!"reached");
	}

	return H_MAKE(XREntry, xr);
}

HParsedToken *
act_xrstment(const HParseResult *p, void *u)
{
	XREntry *xr = H_ALLOC(XREntry);

	xr->obj = NULL;
	xr->type = H_FIELD_UINT(0);
	switch (xr->type) {
	case XR_FREE:
		xr->f.next = H_FIELD_UINT(1);
		xr->f.ngen = H_FIELD_UINT(2);
		break;
	case XR_INUSE:
		xr->n.offs = H_FIELD_UINT(1);
		xr->n.gen = H_FIELD_UINT(2);
		break;
	case XR_OBJSTM:
		xr->o.stm = H_FIELD_UINT(1);
		xr->o.idx = H_FIELD_UINT(2);
		break;
	default:
		assert(!"reached");
	}

	return H_MAKE(XREntry, xr);
}

#define act_xrefs h_act_last

/*
 * return a cross-reference stream in the same form as xref sections.
 *
 * p = ((pnat nat) (dict [HParseResult: xrefs]))
 * result = (xrefs dict)
 */
HParsedToken *
act_xrstm(const HParseResult *p, void *u)
{
	const HParsedToken *xrefs, *dict;
	HParsedToken *tok;
	HParseResult *res;

	dict = H_INDEX_TOKEN(p->ast, 1, 0);
	res = H_FIELD(HParseResult, 1, 1);	// XXX free this
	xrefs = res->ast;

	tok = H_MAKE_SEQN(2);
	tok->seq->elements[0] = (HParsedToken *)xrefs;
	tok->seq->elements[1] = (HParsedToken *)dict;
	tok->seq->used = 2;
	return tok;
}

/*
 * validate the /Type field on a cross-reference stream.
 *
 * p = ((pnat nat) (dict offs offs))
 */
bool
validate_xrstm(HParseResult *p, void *u)
{
	const Dict *tdict = H_FIELD(Dict, 1, 0);
	const HParsedToken *v = dictentry(tdict, "Type");

#if 0
	if (v == NULL)
		fprintf(stderr, "stream dict has no /Type\n");
	else if (v->token_type != TT_BYTES)
		fprintf(stderr, "stream /Type is no name object\n");
	else if (bytes_eq(v->bytes, "XRef"))
		return true;
	return false;
#endif

	return (v != NULL && v->token_type == TT_BYTES &&
	    bytes_eq(v->bytes, "XRef"));
}

HParsedToken *
act_dict_(const HParseResult *p, void *env)
{
	Dict *dict = H_CAST_SEQ(p->ast);

	return H_MAKE(Dict, dict);
}

#define act_array_ h_act_flatten


/*
 * input grammar
 */

HParser *p_pdf;
HParser *p_pdfdbg;
HParser *p_startxref;
HParser *p_xref;
HParser *p_objdef;

/* continuations for h_bind() */
HParser *kstream(HAllocator *, const HParsedToken *, void *);
HParser *kxstream(HAllocator *, const HParsedToken *, void *);

void
init_parser(struct Env *aux)
{
	TT_HParseResult = h_allocate_token_new("HParseResult", NULL, pp_parseresult);
	TT_XREntry =	h_allocate_token_new("XREntry", NULL, pp_xrentry);
	TT_Ref =	h_allocate_token_new("Ref", NULL, pp_ref);
	TT_Dict =	h_allocate_token_new("Dict", NULL, pp_dict);

	/* lines */
	H_RULE(cr,	p_mapch('\r', '\n'));	/* semantic value: \n */
	H_RULE(lf,	h_ch('\n'));		/* semantic value: \n */
	H_RULE(crlf,	h_right(cr, lf));	/* semantic value: \n */
	H_RULE(eol,	CHX(crlf, cr, lf));
	H_RULE(end,	h_end_p());

	/* character classes */
#define LWCHARS	"\0\t\f "
#define WCHARS	LWCHARS "\n\r"
#define DCHARS	"()<>[]{}/%"
	H_RULE(wchar,	IN(WCHARS));			/* white-space */
	H_RULE(lwchar,	IN(LWCHARS));			/* "line" whitespace */
	//H_RULE(dchar,	IN(DCHARS));			/* delimiter */
	H_RULE(rchar,	NOT_IN(WCHARS DCHARS));		/* regular */
	H_RULE(nchar,	NOT_IN(WCHARS DCHARS "#"));	/* name */
	H_RULE(schar,	NOT_IN("()\n\r\\"));		/* string literal */
	H_ARULE(digit,	h_ch_range('0', '9'));
	H_ARULE(pdigit,	h_ch_range('1', '9'));
	H_ARULE(hlower,	h_ch_range('a', 'f'));
	H_ARULE(hupper,	h_ch_range('A', 'F'));
	H_RULE(hdigit,	CHX(digit, hlower, hupper));
	H_ARULE(odigit,	h_ch_range('0', '7'));

	H_RULE(sp,	h_ch(' '));
	H_RULE(percent,	h_ch('%'));
	H_RULE(period,	h_ch('.'));
	H_RULE(slash,	h_ch('/'));
	H_RULE(hash,	h_ch('#'));
	H_RULE(bslash,	h_ch('\\'));
	H_RULE(lparen,	h_ch('('));
	H_RULE(rparen,	h_ch(')'));
	H_RULE(langle,	h_ch('<'));
	H_RULE(rangle,	h_ch('>'));
	H_RULE(lbrack,	h_ch('['));
	H_RULE(rbrack,	h_ch(']'));
	H_RULE(plus,	h_ch('+'));
	H_RULE(minus,	h_ch('-'));

	/* whitespace */
	H_RULE(comment,	SEQ(percent, h_many(NOT_IN("\r\n")), CHX(cr,lf,end)));
		/* ^ NB: must consume cr/lf for ws to be LL(k) */
	H_RULE(wel,	IGN(CHX(wchar, comment)));
	H_RULE(ws,	IGN(h_many(wel)));
	H_RULE(lws,	IGN(h_many(IGN(lwchar))));

	/* tokenization */
#define MANY_WS(X)	h_many(CHX(wel, X))
#define TOK(X)	h_middle(ws, X, h_not(rchar))
#define KW(S)	IGN(TOK(LIT(S)))

	/* misc */
	H_RULE(nl,	IGN(h_right(lws, eol)));
	H_RULE(epsilon,	h_epsilon_p());
	H_RULE(empty,	SEQ(epsilon));
	H_RULE(digits,	h_many1(digit));
	H_ARULE(nat,	digits);
	H_VRULE(pnat,	nat);
	H_RULE(npair,	SEQ(pnat, wel,ws, nat));

	/*
	 * objects
	 */
	
	H_ARULE(ref,	SEQ(npair, wel,ws, LIT("R")));
	H_RULE(null,	LIT("null"));
	H_RULE(boole,	CHX(LIT("true"), LIT("false")));

	/* numbers */
	H_ARULE(sign,	CHX(minus, IGN(plus)));
	H_VRULE(intnn,	nat);
	H_ARULE(realnn,	CHX(SEQ(digits, period, digits),	/* 12.3 */
			    SEQ(digits, period, empty),		/* 123. */
			    SEQ(empty, period, digits)));	/* .123 */
		// XXX ^ we _could_ move the "123." case into intnn...
	H_RULE(numbnn,	CHX(realnn, intnn));
	H_RULE(snumb,	SEQ(sign, numbnn));
	H_ARULE(numb,	CHX(snumb, numbnn));

	/* names */
	H_ARULE(nesc,	SEQ(hash, hdigit, hdigit));
	H_ARULE(nstr,	h_many(CHX(nchar, nesc)));	/* '/' is valid */
	H_RULE(name,	h_right(slash, nstr));

	/* strings */
	H_RULE(snest,	h_indirect());
	H_RULE(bsn,	p_mapch('n', 0x0a));	/* LF */
	H_RULE(bsr,	p_mapch('r', 0x0d));	/* CR */
	H_RULE(bst,	p_mapch('t', 0x09));	/* HT */
	H_RULE(bsb,	p_mapch('b', 0x08));	/* BS (backspace) */
	H_RULE(bsf,	p_mapch('f', 0x0c));	/* FF */
	H_RULE(escape,	CHX(bsn, bsr, bst, bsb, bsf, lparen, rparen, bslash));
	H_ARULE(octal,	CHX(REP(odigit,3), REP(odigit,2), REP(odigit,1)));
	H_RULE(wrap,	IGN(eol));
	H_RULE(sesc,	h_right(bslash, CHX(escape, octal, wrap, epsilon)));
		/* NB: lone backslashes and escaped newlines are ignored */
	H_ARULE(schars,	h_many(CHX(schar, snest, sesc, eol)));
	H_RULE(snest_,	SEQ(lparen, schars, rparen));
	H_RULE(litstr,	h_middle(lparen, schars, rparen));
	H_RULE(hexstr,	h_middle(langle, MANY_WS(hdigit), rangle));
	H_ARULE(string,	CHX(litstr, hexstr));
	h_bind_indirect(snest, snest_);

	H_RULE(array,	h_indirect());
	H_RULE(dict,	h_indirect());

	/* classify objects by whether they start/end with a delimiter: */
	H_RULE(robj,	CHX(ref, null, boole, numb));		/* rchars */
	H_RULE(dobj,	CHX(string, array, dict));		/* dchars */
	H_RULE(obj,	CHX(robj, name, dobj));

	/* dictionaries */
	H_RULE(dopen,	LIT("<<"));
	H_RULE(dclose,	LIT(">>"));
	H_RULE(k_v,	CHX(CHX(SEQ(name, wel,ws, obj),
				SEQ(name, CHX(name,dobj))),
			    VIOL(SEQ(name, wel,ws), "Key with no value (severity=2)")));
	H_ARULE(dict_,	h_middle(dopen, MANY_WS(k_v), dclose));
		// XXX this allows, for instance, "<<<<" to be parsed as "<< <<". ok?
		// XXX validate: dict keys must be unique
	h_bind_indirect(dict, dict_);

	/* arrays */
	H_RULE(elemd,	h_indirect());	/* elements following a delimiter */
	H_RULE(elemr,	h_indirect());	/* elements following a regular char */
	H_ARULE(array_,	h_middle(lbrack, elemd, rbrack));
	H_RULE(elemd_,	CHX(SEQ(ws, dobj, elemd),
			    SEQ(ws, name, elemr),
			    SEQ(ws, robj, elemr),
			    ws));
	H_RULE(elemr_,	CHX(SEQ(ws, dobj, elemd),
			    SEQ(ws, name, elemr),
			    SEQ(wel,ws, robj, elemr),
			    ws));
	h_bind_indirect(elemd, elemd_);
	h_bind_indirect(elemr, elemr_);
	h_bind_indirect(array, array_);

	/* streams */
	H_RULE(stmbeg,	SEQ(dict, ws, LIT("stream"), OPT(cr), lf));
	H_RULE(stmend,	SEQ(OPT(eol), LIT("endstream")));
	H_RULE(stream,	h_left(h_bind(stmbeg, kstream, aux), stmend));
		// XXX is whitespace allowed between the eol and "endstream"?

	/*
	 * file structure
	 */

	/* header */
	H_RULE(version,	SEQ(pdigit, IGN(period), pdigit));
	H_RULE(header,	h_middle(LIT("%PDF-"), version, nl));

	/* body */
	H_RULE(indobj,	CHX(stream, obj));
	H_RULE(objdef,	SEQ(ws, npair, wel, KW("obj"), ws, indobj, KW("endobj")));
	H_RULE(body,	h_many(objdef));

	/* for object streams */
	//H_RULE(osidx,	h_sepBy(npair, SEQ(wel,ws)));

	/* cross-reference section */
	H_RULE(xreol,	CHX(SEQ(sp, cr), SEQ(sp, lf), crlf));
		// ^ XXX does the real world follow this rule?! cf. loop.pdf
	H_RULE(xrtyp,	CHX(h_ch('n'), h_ch('f')));
	H_ARULE(xroff,	REP(digit, 10));
	H_ARULE(xrgen,	REP(digit, 5));
	H_ARULE(xrent,	SEQ(xroff, IGN(sp), xrgen, IGN(sp), xrtyp, IGN(xreol)));
	H_RULE(xrhead,	SEQ(nat, IGN(sp), nat, nl));
	H_RULE(xrsub,	SEQ(xrhead, h_many(xrent)));
	H_ARULE(xrefs,	SEQ(KW("xref"), nl, h_many(xrsub)));

	/* cross-reference streams */
	H_RULE(xstream,	h_bind(stmbeg, kxstream, aux));
	H_AVRULE(xrstm,	SEQ(ws, npair, wel, KW("obj"), ws, xstream));
		// XXX skip however much we consumed and check for "endstream endobj"?

	/* trailer */
	H_RULE(startxr, SEQ(nl, KW("startxref"), nl,
			    lws, nat, nl,
			    LIT("%%EOF"), OPT(nl)));

	/* used for the backwards search */
	H_RULE(lasteof, SEQ(nl, KW("startxref"), nl,
			    lws, nat, nl,
		// XXX the real world sometimes omits nl after %%EOF inside the file.
		//     the next 'tail' would be appended right after the 'F',
		//     presumably because the previous version of the file
		//     ended without a trailing newline. m)
		//     this is invalid per spec, because it creates a run-on 
		//     comment, but we should probably accept-and-warn.
		// XXX should lws be allowed before EOF marker?
		// NB: lws before xref offset is allowed, cf. p.48 (example 4)
			    LIT("%%EOF"),
			    CHX(VIOL(SEQ(nl, h_many1(nl), end),
				     "(offset FROM END) Multiple newlines after final %%EOF (severity=4)"),
				SEQ(h_many(nl), end),
				VIOL(SEQ(h_butnot(h_ch_range(0, 255), LIT("%%EOF"))),
				     "(offset FROM END) Data after final  %%EOF (severity=7)"))));
	
	H_RULE(xr_td,	SEQ(xrefs, KW("trailer"), ws, dict));

	H_RULE(hdr_junk,    VIOL(h_many1(h_butnot(h_ch_range(0, 255), objdef)),
				 "Uncommented junk after header (severity=1)"));
	H_RULE(tail,	SEQ(body, h_optional(xr_td), startxr));
		// XXX the real world likes to omit 'startxr' from all but the
		// last trailer. we should accept-and-warn in that case.
	H_RULE(final_eof_junk, CHX(VIOL(SEQ(h_many1(nl), end), "Multiple newlines after final %%EOF (severity=4)"),
				   VIOL(h_many1(h_butnot(h_ch_range(0, 255), LIT("%%EOF"))),
					"Data after final %%EOF (severity=7)"),
				   end));
	H_RULE(pdf,	SEQ(header, OPT(hdr_junk), h_many1(tail), final_eof_junk));

	/* debug parser to consume as much as possible */
	H_RULE(pdfdbg,	SEQ(header, OPT(hdr_junk), h_many(tail), body, OPT(xr_td), OPT(startxr), final_eof_junk));


	/* global parser variables */
	p_pdf = pdf;
	p_pdfdbg = pdfdbg;
	p_startxref = lasteof; //startxr;
	p_xref = CHX(xr_td, xrstm);
	p_objdef = objdef;

	p_fail = h_nothing_p();
	p_epsilon = epsilon;
	p_return_0 = h_action(epsilon, act_return_uint, (void *)0);
	p_return_1 = h_action(epsilon, act_return_uint, (void *)1);
}


/*
 * lookup and resolution of indirect references
 */

XREntry *
lookup_xref(struct Env *aux, size_t nr, size_t gen)
{
	HParsedToken *ss;	/* xref subsection */
	size_t base, n;

	/* for each cross-reference section (i.e. update) */
	for (size_t i = 0; i < aux->nxrefs; i++) {
		HCountedArray *subs = H_INDEX_SEQ(aux->xrefs[i], 0);

		/* for each cross-reference subsection */
		for (size_t j = 0; j < subs->used; j++) {
			ss = subs->elements[j];
			base = H_INDEX_UINT(ss, 0, 0);
			n = H_INDEX_UINT(ss, 0, 1);

			if (nr >= base && nr - base < n)
				return H_INDEX(XREntry, ss, 1, nr - base);
		}
	}

	return NULL;
}

const HParsedToken *
parse_obj(struct Env *aux, size_t nr, size_t gen, size_t offset)
{
	HParseResult *res;
	size_t def_nr, def_gen;

	if (offset >= aux->sz) {
		fprintf(stderr, "%s: position %zu (%#zx) for object %zu %zu is "
		    "out of bounds\n", aux->infile, offset, offset, nr, gen);
		return NULL;
	}

	res = h_parse(p_objdef, aux->input + offset, aux->sz - offset);
	if (res == NULL) {
		fprintf(stderr, "%s: error parsing object %zu %zu at position "
		    "%zu (%#zx)\n", aux->infile, nr, gen, offset, offset);
		return NULL;
	}
	assert(res->ast != NULL && res->ast->token_type == TT_SEQUENCE);
	/* res->ast = ((nr gen) obj) */

	def_nr = H_INDEX_UINT(res->ast, 0, 0);
	def_gen = H_INDEX_UINT(res->ast, 0, 1);
	if (def_nr != nr || def_gen != gen) {
		fprintf(stderr, "%s: object ID mismatch at position %zu "
		    "(%#zx): sought %zu %zu, found %zu %zu.\n", aux->infile,
		    offset, offset, nr, gen, def_nr, def_gen);
		return NULL;
	}

	return H_INDEX_TOKEN(res->ast, 1);
}

const HParsedToken *
parse_objstm_obj(struct Env *aux, size_t nr, size_t stm_nr, size_t idx)
{
	XREntry *ent;
	const HParsedToken *stm;

	/*
	 * acquire the stream object
	 */

	ent = lookup_xref(aux, stm_nr, 0);
	if (ent == NULL)
		return NULL;		/* stream not found */

	switch (ent->type)
	{
	case XR_FREE:
		return NULL;		/* stream deleted */
	case XR_INUSE:
		if (ent->n.gen != 0)
			return NULL;	/* stream replaced */
		if (ent->obj == NULL)
			ent->obj = parse_obj(aux, stm_nr, 0, ent->n.offs);
		break;
	case XR_OBJSTM:
		return NULL;		/* invalid: nested streams */
	}

	if ((stm = ent->obj) == NULL) {
		fprintf(stderr, "%s: error parsing object stream at position "
		    "%zu (%#zx)\n", aux->infile, ent->n.offs, ent->n.offs);
		return NULL;
	}

	/*
	 * decode the stream and find the target object in it
	 */
	return NULL;	// XXX
}

const HParsedToken *
resolve(struct Env *aux, const HParsedToken *v)
{
	XREntry *ent = NULL;
	Ref *r;

	/* direct objects pass through */
	if (v == NULL || v->token_type != TT_Ref)
		return v;

	/* we are looking at an indirect reference */
	r = v->user;

	/* find the xref entry for this reference */
	ent = lookup_xref(aux, r->nr, r->gen);
	if (ent == NULL)
		return NULL;			/* obj not found */
	if (ent->obj != NULL)
		return resolve(aux, ent->obj);

	/* parse the object and memoize */
	ent->obj = v;				/* break loops */
	switch (ent->type)
	{
	case XR_FREE:
		return NULL;			/* obj deleted */
	case XR_INUSE:
		if (ent->n.gen != r->gen)
			return NULL;		/* obj nr reused */
		ent->obj = parse_obj(aux, r->nr, r->gen, ent->n.offs);
		break;
	case XR_OBJSTM:
		if (r->gen != 0)
			return NULL;		/* invalid entry! */
		ent->obj = parse_objstm_obj(aux, r->nr, ent->o.stm, ent->o.idx);
		break;
	}

	return resolve(aux, ent->obj);
}


/*
 * stream object handling incl. filters and cross-reference streams
 */

#include <limits.h>	/* INT_MAX */
#include <stdlib.h>	/* abs() */
#include <zlib.h>
#include <err.h>

struct predictor {
	/* parameters */
	int num;	/* default: 1 (no prediction) */
	int colors;	/* default: 1 */
	int bpc;	/* bits per component; default: 8 */
	int columns;	/* default: 1 */

	int rowsz;	/* bytes per row = ceil(colors * bpc * columns / 8) */

	/* state */
	HSuspendedParser *sp;
	uint8_t	(*predfun)(int, int, int);
	uint8_t	*buf;		/* previous row of input */
	uint8_t	 c;		/* byte 'c' (upper left) */
	int	 x;		/* current position */
};

int
depred_none(struct predictor *pred, uint8_t *inp, size_t sz)
{
	return h_parse_chunk(pred->sp, inp, sz);
}

uint8_t pp_none(int a, int b, int c)	{ return 0; }
uint8_t pp_sub(int a, int b, int c)	{ return a; }
uint8_t pp_up(int a, int b, int c)	{ return b; }
uint8_t pp_avg(int a, int b, int c)	{ return (a + b) / 2; }

uint8_t
pp_paeth(int a, int b, int c)
{
	int p	= a + b - c;
	int pa	= abs(p - a);
	int pb	= abs(p - b);
	int pc	= abs(p - c);

	if (pa <= pb && pa <= pc) return a;
	if (pb <= pc) return b;
	return c;
}

int
depred_png(struct predictor *pred, uint8_t *inp, size_t sz)
{
	/* NB:
	 * at this point, the specific value of pred->num no longer matters.
	 * the PNG predictor tags each row with the function used for that row
	 * and decoding always follows the tag.
	 */
	static uint8_t (*predfuns[])(int, int, int) =
	    {pp_none, pp_sub, pp_up, pp_avg, pp_paeth};

	bool done = false;
	int bpp;

	bpp = (pred->colors * pred->bpc + 7) / 8;	/* bytes per pixel */
	assert (bpp > 0);

	for (size_t i=0; i < sz && !done; i++) {
		int x = pred->x;
		int a = x<bpp ? 0 : pred->buf[x-bpp];	/* left */
		int b = pred->buf[x];			/* up */
		int c = pred->c;			/* up left */

		if (pred->predfun == NULL) {	/* we are before a new row */
			/* select predictor function */
			if (inp[i] > 4) {
				fprintf(stderr, "unknown PNG predictor %d\n",
				    (int)inp[i]);
				return -1;
			}
			pred->predfun = predfuns[inp[i]];

			/* consume the tag */
			if (++i == sz)
				break;
		}

		/* undo the prediction and save the decoded value */
		pred->buf[x] = inp[i] + pred->predfun(a, b, c);

		/* advance to the right */
		pred->c = b;
		pred->x = ++x;

		/* when row complete, pass it to parser and start a new row */
		if (x == pred->rowsz) {
			done = h_parse_chunk(pred->sp, pred->buf, pred->rowsz);
			pred->c = pred->x = 0;
			if (pred->num != 2)	/* support for 8-bpc TIFF */
				pred->predfun = NULL;
		}
	}

	return done;
}

HParseResult *
FlateDecode(const Dict *parms, HBytes b, HParser *p)
{
	size_t const BUFSIZE = 8 * 1024;
	uint8_t *buf;
	HSuspendedParser *sp;
	HParseResult *res;
	const HParsedToken *v;
	size_t sz;
	int done;
	z_stream strm = {0};
	int ret;
	struct predictor pred = {1, 1, 8, 1};
	int (*depredict)(struct predictor *, uint8_t *, size_t);

	/* set up the predictor (if any) */
	#define SETPARM(VAR,STR) do {					\
		v = dictentry(parms, (STR));				\
		if (v != NULL) {					\
			if (v->token_type != TT_SINT || v->sint < 0)	\
				return NULL;				\
			VAR = v->sint;					\
		} } while(0)
	SETPARM(pred.num,	"Predictor");
	SETPARM(pred.colors,	"Colors");
	SETPARM(pred.bpc,	"BitsPerComponent");
	SETPARM(pred.columns,	"Columns");
	#undef SETPARM
	if (pred.num == 1)
		depredict = depred_none;
	else {
		if (pred.num >= 10 && pred.num <= 15)
			depredict = depred_png;
		else if (pred.num == 2) {
			/* for 8-bpc TIFF pred. 2, we can reuse PNG Sub */
			if (pred.bpc == 8) {
				pred.predfun = pp_sub;	/* predict left */
				depredict = depred_png;
			} else {
				// XXX add general TIFF predictor (bpc != 8)
				fprintf(stderr, "FlateDecode: /Predictor %d "
				    "not supported for /BitsPerComponent %d\n",
				    pred.num, pred.bpc);
				return NULL;
			}
		} else {
			fprintf(stderr, "FlateDecode: /Predictor %d"
			    " not supported\n", pred.num);
			return NULL;
		}

		/* allocate row buffer */
		if (pred.columns > (INT_MAX - 7) / pred.colors / pred.bpc) {
			fprintf(stderr, "FlateDecode: overflow\n");
			return NULL;
		}
		pred.rowsz = (pred.colors * pred.bpc * pred.columns + 7) / 8;
		pred.buf = calloc(1, pred.rowsz);
		if (pred.buf == NULL)
			err(1, "FlateDecode");
	}

	/* set up zlib */
	// XXX pass our allocator to zlib
	ret = inflateInit(&strm);
	if (ret != Z_OK)
		errx(1, "inflateInit: %s (%d)", strm.msg, ret);
	buf = malloc(BUFSIZE);
	if (buf == NULL)
		err(1, "FlateDecode");

	/* initialize target parser */
	sp = h_parse_start(p);
	assert(sp != NULL);
	pred.sp = sp;

	done = 0;
	strm.avail_in = b.len;
	strm.next_in = (unsigned char *)b.token;
	do {
		strm.avail_out = BUFSIZE;
		strm.next_out = buf;

		ret = inflate(&strm, Z_NO_FLUSH);
		if (ret != Z_STREAM_END && ret != Z_OK) {
			fprintf(stderr, "inflate: %s (%d)\n", strm.msg, ret);
			break;
		}

		sz = BUFSIZE - strm.avail_out;
		done = depredict(&pred, buf, sz);
	} while (done == 0 && ret == Z_OK);

	res = h_parse_finish(sp);
		// XXX always return NULL on error?
	inflateEnd(&strm);
	free(pred.buf);
	free(buf);

	if (done == -1)
		return NULL;
	return res;
}

/*
 * decode the bytes in 'b' according to metadata in the stream dictionary 'd'
 * and parse the result with 'p'.
 */
HParseResult *
decode_stream(const Dict *d, HBytes b, HParser *p)
{
	HParseResult *(*filter)(const Dict *, HBytes, HParser *);
	const Dict *parms = NULL;
	const HParsedToken *v;

	v = dictentry(d, "Filter");
	if (v == NULL)
		return h_parse(p, b.token, b.len);

	/* compile to a CF backend to enable incremental parsing */
	if (h_compile(p, PB_LLk, NULL) == -1)
		errx(1, "stream data parser: LL(1) compile failed");

	if (v->token_type == TT_SEQUENCE)
		return NULL;	// XXX filter chains not supported, yet
	assert(v->token_type == TT_BYTES);
	if (bytes_eq(v->bytes, "FlateDecode"))
		filter = FlateDecode;
	else
		return NULL;		/* filter not supported */

	v = dictentry(d, "DecodeParms");
	if (v && v->token_type == TT_Dict)
		parms = v->user;

	return filter(parms, b, p);
}

HParsedToken *
act_rest(const HParseResult *p, void *env)
{
	struct Env *aux = env;
	size_t offset = H_CAST_UINT(p->ast) / 8;

	return H_MAKE_BYTES(aux->input + offset, aux->sz - offset);
}

HParser *
p_rest__m(HAllocator *mm__, struct Env *aux)
{
	return h_action__m(mm__, h_tell__m(mm__), act_rest, aux);
}

/* combine current position with env=(input,sz) into HBytes */
HParsedToken *
act_take_bytes(const HParseResult *p, void *env)
{
	const HBytes *bs = env;
	size_t offset = H_CAST_UINT(p->ast) / 8;

	/*
	 * NB: we must allocate a new HBytes struct here because the old one is
	 * allocated only temporarily for the lifetime of the continuation
	 * below.
	 */
	return H_MAKE_BYTES(bs->token + offset, bs->len);
}

HParser *
p_take__m(HAllocator *mm__, size_t n, struct Env *aux)
{
	HParser *skip, *bytes;
	HBytes *bs;

	/* dummy struct to hold the pair (input,n) */
	bs = h_alloc(mm__, sizeof(HBytes));
	bs->token = aux->input;
	bs->len = n;

	bytes = h_action__m(mm__, h_tell__m(mm__), act_take_bytes, bs);
	skip  = h_skip__m(mm__, n * 8);

	return h_left__m(mm__, bytes, skip);
}

HParser *
p_xrefdata__m(HAllocator *mm__, const Dict *dict);

HParser *
p_stream_data__m(HAllocator *mm__, const Dict *dict)
{
	const HParsedToken *v;

	v = dictentry(dict, "Type");
	if (v == NULL || v->token_type != TT_BYTES)	// XXX -> custom type
		return NULL;				/* no /Type field */

	/* interpret known stream types */
	if (bytes_eq(v->bytes, "XRef"))
		return p_xrefdata__m(mm__, dict);
	// XXX
	//if (bytes_eq(v->bytes, "ObjStm"))
	//	return p_objstm__m(mm__, dict);

	return NULL;					/* unrecognized type */
}

struct streamspec {
	Dict *dict;		/* stream dictionary */
	HParser *parser;	/* data parser */
};

HParsedToken *
act_ks_value(const HParseResult *p, void *u)
{
	struct streamspec *spec = u;
	HBytes bytes = H_CAST_BYTES(p->ast);
	HParseResult *res;

	/* decode and parse the stream data */
	res = decode_stream(spec->dict, bytes, spec->parser);
	assert(res != NULL);	// XXX parse failure!

	return H_MAKE(HParseResult, res);
}

/*
 * This continuation takes the stream dictionary (as first element of x) and
 * should return a parser that consumes exactly the bytes that make up the
 * stream data.
 *
 * x = (dict ...)
 */
HParser *
kstream(HAllocator *mm__, const HParsedToken *x, void *env)
{
	struct Env *aux = env;
	HParsedToken *dict_t = H_INDEX_TOKEN(x, 0);
	Dict *dict = H_CAST(Dict, dict_t);
	const HParsedToken *v = NULL;
	HParser *bytes_p, *dict_p, *value_p;
	struct streamspec *spec;
	size_t sz;

	/* look for the Length entry */
	v = dictentry(dict, "Length");
	v = resolve(aux, v);		/* resolve indirect references */
	if (v == NULL || v->token_type != TT_SINT || v->sint < 0)
		goto fail;
	sz = (size_t)v->sint;

	//fprintf(stderr, "parsing stream object, length %zu.\n", sz);	// XXX debug

	dict_p	= p_return__m(mm__, dict_t);
	bytes_p = p_take__m(mm__, sz, aux);

	spec = h_alloc(mm__, sizeof(struct streamspec));
	spec->dict = dict;
	spec->parser = p_stream_data__m(mm__, dict);
	if (spec->parser != NULL)
		value_p = h_action__m(mm__, bytes_p, act_ks_value, spec);
	else
		value_p = bytes_p;

	return h_sequence__m(mm__, dict_p, value_p, NULL);
fail:
#if 0
	if (v == NULL)
		fprintf(stderr, "stream /Length missing\n");
	else if (v -> token_type != TT_SINT)
		fprintf(stderr, "stream /Length not an integer\n");
	else if (v < 0)
		fprintf(stderr, "stream /Length negative\n");
#endif
	//h_pprintln(stderr, p);	// XXX debug
	return p_fail;
}

HParser *
p_xrefsub__m(HAllocator *mm__, size_t base, size_t count, HParser *p_entry)
{
	HParser *ret_base, *ret_count, *p_header, *p_entries;

	ret_base  = p_return_uint__m(mm__, base);
	ret_count = p_return_uint__m(mm__, count);
	p_header  = h_sequence__m(mm__, ret_base, ret_count, NULL);
	p_entries = h_repeat_n__m(mm__, p_entry, count);

	return h_sequence__m(mm__, p_header, p_entries, NULL);
}

HParser *
p_xrefdata__m(HAllocator *mm__, const Dict *dict)
{
	const HParsedToken *v;
	HParser *p_field[3], *p_entry, **p_subs;
	size_t W[3];
	size_t Size, Wn, Wskip;

	/*
	 * what follows is a horrible bunch of code that builds, from the
	 * entries /W, /Index, and /Size in the stream dictionary, a parser for
	 * the cross-reference data itself.
	 *
	 * in short, every cross-reference entry consists of (as of PDF 2.0)
	 * three fields, but it could be more. /W gives the widths (in bytes)
	 * of these fields. the /Index specifies the division of the data into
	 * subsections; it is an array of natural numbers that in pairs specify
	 * the base object number and length of each subsection - analogous to
	 * the subsection headers in classic xref sections.
	 *
	 * when /Index is missing, a default value of [0 Size] is defined,
	 * where Size is the value of the /Size field. as in normal trailer
	 * dictionaries, it specifies the total size of the (entire)
	 * cross-reference table.
	 *
	 * when /W states a width of 0 for a field, that field is not present
	 * in the data and a default value should be used "if there is one".
	 * most notably, the first field determines the "type" of the entry,
	 * analogous to the 'n' and 'f' tags in classic xref sections; a width
	 * of 0 for the first field is specified to mean that every entry is of
	 * type 1 (= "n"). that type, in particular, specifies a default of 0
	 * for field 3 (generation). in fact, these are the only defaults
	 * defined by ISO 32000-1:2008 (PDF 1.7).
	 *
	 *   entry type	 field no.  default value
	 *		 1 (type)   1
	 *   1 ("n")	 3 (gen.)   0
	 */

	/* Size (required) - total size of xref table */
	v = dictentry(dict, "Size");
	if (v == NULL || v->token_type != TT_SINT || v->sint < 1)
		return p_fail;
	Size = v->sint;

	/* W (required) - field widths for each xref entry */
	v = dictentry(dict, "W");
	if (v == NULL || v->token_type != TT_SEQUENCE)
		return p_fail;
	if ((Wn = v->seq->used) < 3)
		return p_fail;
	Wskip = 0;
	for (size_t i = 0; i < Wn; i++) {
		HTokenType tt = v->seq->elements[i]->token_type;
		int64_t w = v->seq->elements[i]->sint;

		if (tt != TT_SINT || w < 0)
			return p_fail;
		if (i < 3) {
			/* we can't take >64 bits and want to use size_t */
			if (w > 8 || (uint64_t)w > sizeof(size_t))
				return p_fail;
			W[i] = (size_t)w;
		} else {
			if ((uint64_t)w > SIZE_MAX - Wskip)
				return p_fail;	/* overflow */
			Wskip += w;
		}
	}
	if (Wskip > SIZE_MAX / 8)
		return p_fail;

	/*
	 * build the parser for one xref entry.
	 *
	 * in summary, the only sensible forms for /W are:
	 *
	 *   [t x y] with t,x,y > 0  full general form
	 *   [0 x y] with x,y > 0    only type-1 ("in use") entries
	 *   [0 x 0] with x > 0	     only type-1 entries, only offsets
	 *
	 * however, though nonsensical, [t x 0] with t,x > 0 is not disallowed
	 * by the spec; as long as all entries are of type 1, the xref data can
	 * be interpreted without ambiguity.
	 *
	 * in fact, every nonsensical form is possible as long as there are 0
	 * entries.
	 *
	 * we realize this mess by just initializing the default parser to
	 * p_fail and and replacing the known cases afterwards.
	 */
	for (size_t i = 0; i < 3; i++) {
		if (W[i] == 0)
			p_field[i] = p_fail;	/* no known default */
		else
			p_field[i] = h_bits__m(mm__, W[i] * 8, false);
	}
	/* known default cases: */
	if (W[0] == 0)
		p_field[0] = p_return_1;	/* all type 1 */
	if (W[2] == 0) {
		p_field[2] = p_return_0;	/* all generation 0 */
		#if 0
		/* XXX
		 * i've seen a 0-width field 3 used with values of 1 (inuse)
		 * and 2 (objstm) in field 1, implying "objstm idx 0" for the
		 * latter case.
		 */
		if (W[0] > 0) {
			/* type field *must* be 1 */
			p_field[0] = h_attr_bool__m(mm__, p_field[0],
			    validate_eq_uint, (void *)1);
		}
		#endif
	}
	if (Wskip > 0)	// XXX h_skip does not work with CF, yet
		return p_fail;
	p_entry = h_sequence__m(mm__, p_field[0], p_field[1], p_field[2], NULL);
	p_entry = h_action__m(mm__, p_entry, act_xrstment, NULL);

	/* Index (optional) - subsections [base count ...] */
	v = dictentry(dict, "Index");
	if (v == NULL) {
		/* default: [0 Size] */
		p_subs = h_alloc(mm__, 2 * sizeof(HParser *));
		p_subs[0] = p_xrefsub__m(mm__, 0, Size, p_entry);
		p_subs[1] = NULL;
	} else if (v->token_type != TT_SEQUENCE) {
		return p_fail;
	} else {
		size_t nsubs = v->seq->used / 2;

		/* build a parser for each subsection */
		if (nsubs >= SIZE_MAX / sizeof(HParser *))
			return p_fail;
		p_subs = h_alloc(mm__, (nsubs + 1) * sizeof(HParser *));
		for (size_t i = 0; i < nsubs; i++) {
			HParsedToken *base = v->seq->elements[2 * i];
			HParsedToken *n = v->seq->elements[2 * i + 1];

			if (base->token_type != TT_SINT || base->sint < 0 ||
			    n->token_type != TT_SINT || n->sint < 0 ||
			    (uint64_t)n->sint > SIZE_MAX)
				return p_fail;

			p_subs[i] = p_xrefsub__m(mm__, base->sint, n->sint,
			    p_entry);
		}
		p_subs[nsubs] = NULL;
	}
	return h_sequence__ma(mm__, (void **)p_subs);
}

/*
 * This continuation is very similar to kstream, except that it does not
 * rely on /Length to consume the right amount of input. If /Length is
 * not present or indirect, it will operate on the entire rest of the input.
 * This is permissible, other than for general streams, because the XRef data
 * is always self-delimiting.
 *
 * x = (dict ...)
 */
HParser *
kxstream(HAllocator *mm__, const HParsedToken *x, void *env)
{
	struct Env *aux = env;
	HParsedToken *dict_t = H_INDEX_TOKEN(x, 0);
	Dict *dict = H_CAST(Dict, dict_t);
	const HParsedToken *v;
	HParser *bytes_p, *dict_p, *value_p;
	struct streamspec *spec;

	/* restrict bytes to Length if present (and not indirect) */
	v = dictentry(dict, "Length");
	if (v != NULL && v->token_type == TT_SINT && v->sint >= 0)
		bytes_p = p_take__m(mm__, v->sint, aux);
	else
		bytes_p = p_rest__m(mm__, aux);	// XXX consume the proper amount

	/* construct the parser for the stream data */
	spec = h_alloc(mm__, sizeof(struct streamspec));
	spec->dict = dict;
	spec->parser = p_xrefdata__m(mm__, dict);
	assert (spec->parser != NULL);

	dict_p	= p_return__m(mm__, dict_t);
	value_p = h_action__m(mm__, bytes_p, act_ks_value, spec);

	return h_sequence__m(mm__, dict_p, value_p, NULL);
}


/*
 * main program
 */

#include <stdio.h>
#include <inttypes.h>
#include <stdlib.h>	/* realloc() */
#include <fcntl.h>	/* open() */
#include <unistd.h>	/* lseek() */
#include <sys/mman.h>	/* mmap() */

const char *infile = NULL;

/*
 * This helper implements the standard backwards parsing strategy to read all
 * cross-reference sections and trailer dictionaries, starting from the
 * 'startxref' offset found at the very end of the input.
 *
 * Allocates and returns an array of HParsedTokens, each containing the result
 * of a successful 'p_xref' parse. Sets the output parameter 'nxrefs' to the
 * number of elements.
 *
 * A return value of NULL indicates an empty result.
 */
const HParsedToken **
parse_xrefs(const uint8_t *input, size_t sz, size_t *nxrefs)
{
	HParseResult *res = NULL;
	const HParsedToken **xrefs = NULL;	/* empty result */
	const HParsedToken *tok = NULL;
	size_t n = 0, nfwd = 0;
	size_t offset = 0;

	// XXX try formulating this as a parser using h_seek()

	/* search for the "startxref" section from the back of the file */
	HParser *p = h_left(p_startxref, h_end_p());
	for (size_t i = 0; i < sz; i++) {
		res = h_parse(p, input + sz - i, i);
		if (res != NULL)
			break;
	}
	if (res == NULL) {
		fprintf(stderr, "%s: startxref not found\n", infile);
		goto end;
	}
	offset = H_INDEX_UINT(res->ast, 0);

	for (;;) {
		//res = h_parse(p_xref, input + offset, sz - offset);
		HParser *p = h_right(h_seek(offset * 8, SEEK_SET), p_xref);	// XXX
		res = h_parse(p, input, sz);
		if (res == NULL) {
			fprintf(stderr, "%s: error parsing xref section at "
			    "position %zu (%#zx)\n", infile, offset, offset);
			break;
		}
		assert(res->ast != NULL);

		/* save this section in xrefs */
		if (n >= SIZE_MAX / sizeof(HParsedToken *))
			errx(1, "realloc: size would overflow");
		xrefs = realloc(xrefs, (n + 1) * sizeof(HParsedToken *));
		if (xrefs == NULL)
			err(1, "realloc");
		xrefs[n++] = res->ast;

		/* look up the next offset (to the previous xref section) */
		tok = dictentry(H_INDEX(Dict, res->ast, 1), "Prev");
		if (tok == NULL)
			break;
		if (tok->token_type != TT_SINT) {
			fprintf(stderr, "%s: /Prev not an integer\n", infile);
			break;
		}
		if (tok->sint < 0) {
			fprintf(stderr, "%s: /Prev negative\n", infile);
			break;
		}

		/*
		 * validate the new offset. we don't want to get caught in a
		 * loop. the offsets should strictly decrease, unless the file
		 * is a "linearized" PDF. in that case there should be exactly
		 * one xref section at the beginning of the file that is
		 * allowed to point forward.
		 */
		if ((uint64_t)tok->sint >= offset)
			nfwd++;
		if (nfwd > 1) {
			fprintf(stderr, "%s: /Prev pointer of xref section at "
			    "%zu (%#zx) points forward\n", infile, offset,
			    offset);
			break;
		}

		offset = (size_t)tok->sint;
	}

end:
	*nxrefs = n;
	return xrefs;
}

int
main(int argc, char *argv[])
{
	struct Env aux;
	HParseResult *res = NULL;
	const uint8_t *input;
	off_t sz;
	int fd;

	/* command line handling */
	if (argc > 3) {
		fprintf(stderr, "usage: %s file\n", argv[0]);
		return 1;
	}
	if (argc == 3) {
	  H_RULE(nat, h_action(h_many1(h_action(h_ch_range('0', '9'), act_digit, NULL)),
			       act_nat, NULL));
	  strictness = h_parse(nat, (uint8_t *)argv[2], strlen(argv[2]))->ast->uint;
	}
	infile = argv[1];

	/* mmap the input file */
	fd = open(infile, O_RDONLY);
	if (fd == -1)
		err(1, "%s", infile);
	sz = lseek(fd, 0, SEEK_END);
	if (sz == -1)
		err(1, "lseek");
	assert(sz >= 0);
	input = mmap(NULL, sz?sz:1, PROT_READ, MAP_PRIVATE, fd, 0);
	if (input == MAP_FAILED)
		err(1, "mmap");

	/* build parsers */
	aux = (struct Env){infile, input, sz};
	init_parser(&aux);

	/* parse all cross-reference sections and trailer dictionaries */
	aux.xrefs = parse_xrefs(input, sz, &aux.nxrefs);
	// XXX debug
	//fprintf(stderr, "%s: %zu xref sections parsed\n", infile, aux.nxrefs);
	//for (size_t i = 0; i < aux.nxrefs; i++)
	//	h_pprintln(stderr, aux.xrefs[i]);

	/* run the main parser */
	res = h_parse(p_pdf, input, sz);
	if (!res) {
		fprintf(stderr, "%s: no parse\n", infile);

		/* help us find the error */
		res = h_parse(p_pdfdbg, input, sz);
		if (res) {
			int64_t pos = res->bit_length / 8;
			fprintf(stderr, "%s: error after position"
			    " %" PRId64 " (%#" PRIx64 ")\n",
			    infile, pos, pos);
			//h_pprintln(stderr, res->ast);	// XXX debug
		}

		return 1;
	}

	/* print result */
	h_pprintln(stdout, res->ast);

	return 0;
}
